@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2008-2014 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2008-2015 -- leonerd@leonerd.org.uk
 
 package Net::Async::HTTP;
 
@@ -9,11 +9,12 @@ use strict;
 use warnings;
 use base qw( IO::Async::Notifier );
 
-our $VERSION = '0.38';
+our $VERSION = '0.39';
 
 our $DEFAULT_UA = "Perl + " . __PACKAGE__ . "/$VERSION";
 our $DEFAULT_MAXREDIR = 3;
 our $DEFAULT_MAX_IN_FLIGHT = 4;
+our $DEFAULT_MAX_CONNS_PER_HOST = $ENV{NET_ASYNC_HTTP_MAXCONNS} // 1;
 
 use Carp;
 
@@ -29,8 +30,8 @@ use IO::Async::Loop 0.59; # ->connect( handle ) ==> $stream
 use Future 0.28; # ->set_label
 use Future::Utils 0.16 qw( repeat );
 
-use Scalar::Util qw( blessed );
-use List::Util qw( first );
+use Scalar::Util qw( blessed reftype );
+use List::Util 1.29 qw( first pairs );
 use Socket qw( SOCK_STREAM IPPROTO_IP IP_TOS );
 BEGIN {
    if( $Socket::VERSION >= '2.010' ) {
@@ -147,7 +148,7 @@ sub _init
    $self->{read_len}  = READ_LEN;
    $self->{write_len} = WRITE_LEN;
 
-   $self->{max_connections_per_host} = 1;
+   $self->{max_connections_per_host} = $DEFAULT_MAX_CONNS_PER_HOST;
 
    $self->{ssl_params} = {};
 }
@@ -177,9 +178,18 @@ Give 0 to disable the limit entirely.
 =head2 max_connections_per_host => INT
 
 Optional. Controls the maximum number of connections per hostname/server port
-pair, before requests will be queued awaiting one to be free. If not supplied,
-will default to 1. Give 0 to disable the limit entirely. See also the
-L</Connection Pooling> section documented above.
+pair, before requests will be queued awaiting one to be free. Give 0 to
+disable the limit entirely. See also the L</Connection Pooling> section
+documented above.
+
+Currently, if not supplied it will default to 1. However, it has been found in
+practice that most programs will raise this limit to something higher, perhaps
+3 or 4. Therefore, a future version of this module may set a higher value.
+
+To test if your application will handle this correctly, you can set a
+different default by setting an environment variable:
+
+ $ NET_ASYNC_HTTP_MAXCONNS=3 perl ...
 
 =head2 timeout => NUM
 
@@ -269,6 +279,12 @@ Additionally, any parameters whose names start with C<SSL_> will be stored and
 passed on requests to perform SSL requests. This simplifies configuration of
 common SSL parameters.
 
+=head2 require_SSL => BOOL
+
+Optional. If true, then any attempt to make a request that does not use SSL
+(either by calling C<request>, or as a result of a redirection) will
+immediately fail.
+
 =cut
 
 sub configure
@@ -279,7 +295,7 @@ sub configure
    foreach (qw( user_agent max_redirects max_in_flight max_connections_per_host
       timeout stall_timeout proxy_host proxy_port cookie_jar pipeline
       local_host local_port local_addrs local_addr fail_on_error
-      read_len write_len decode_content ))
+      read_len write_len decode_content require_SSL ))
    {
       $self->{$_} = delete $params{$_} if exists $params{$_};
    }
@@ -377,7 +393,7 @@ sub get_connection
 
    # Have a look to see if there are any idle connected ones first
    foreach my $conn ( @$conns ) {
-      $conn->is_idle and $conn->read_handle and return Future->new->done( $conn );
+      $conn->is_idle and $conn->read_handle and return Future->done( $conn );
    }
 
    my $ready = $args{ready};
@@ -497,6 +513,12 @@ The type of non-form data C<content>.
 Optional. If both are given, the HTTP Basic Authorization header will be sent
 with these details.
 
+=item headers => ARRAY|HASH
+
+Optional. If provided, contains additional HTTP headers to set on the
+constructed request object. If provided as an ARRAY reference, it should
+contain an even-sized list of name/value pairs.
+
 =item proxy_host => STRING
 
 =item proxy_port => INT
@@ -628,6 +650,11 @@ sub _do_one_request
 
    $self->prepare_request( $request );
 
+   if( $self->{require_SSL} and not $args{SSL} ) {
+      return Future->fail( "Non-SSL request is not allowed with 'require_SSL' set",
+         http => undef, $request );
+   }
+
    return $self->get_connection(
       host => $args{proxy_host} || $self->{proxy_host} || $host,
       port => $args{proxy_port} || $self->{proxy_port} || $port,
@@ -692,12 +719,12 @@ sub _do_request
             $location = "http://$hostport" . $location;
          }
          else {
-            return Future->new->fail( "Unrecognised Location: $location", http => $previous_response, $request );
+            return Future->fail( "Unrecognised Location: $location", http => $previous_response, $request );
          }
 
          my $loc_uri = URI->new( $location );
          unless( $loc_uri ) {
-            return Future->new->fail( "Unable to parse '$location' as a URI", http => $previous_response, $request );
+            return Future->fail( "Unable to parse '$location' as a URI", http => $previous_response, $request );
          }
 
          $self->debug_printf( "REDIRECT $loc_uri" );
@@ -754,7 +781,7 @@ sub _do_request
             my $message = $resp->message;
             $message =~ s/\r$//; # HTTP::Message bug
 
-            return Future->new->fail( "$code $message", http => $resp, $request );
+            return Future->fail( "$code $message", http => $resp, $request );
          }
 
          return $f;
@@ -779,7 +806,7 @@ sub do_request
    if( $args{on_header} ) {
       # ok
    }
-   elsif( my $on_response = delete $args{on_response} or defined wantarray ) {
+   elsif( $args{on_response} or defined wantarray ) {
       $args{on_header} = sub {
          my ( $response ) = @_;
          return sub {
@@ -787,7 +814,6 @@ sub do_request
                $response->add_content( @_ );
             }
             else {
-               $on_response->( $response ) if $on_response;
                return $response;
             }
          };
@@ -821,6 +847,13 @@ sub do_request
       $on_error->( $message, @rest );
    }) if $on_error;
 
+   if( my $on_response = delete $args{on_response} ) {
+      $future->on_done( sub {
+         my ( $response ) = @_;
+         $on_response->( $response );
+      });
+   }
+
    # DODGY HACK:
    # In void context we'll lose reference on the ->wait_any Future, so the
    # timeout logic will never happen. So lets purposely create a cycle by
@@ -873,6 +906,14 @@ sub _make_request_for_uri
    $request->protocol( "HTTP/1.1" );
    $request->header( Host => $uri->host );
 
+   my $headers = $args{headers};
+   if( $headers and reftype $headers eq "ARRAY" ) {
+      $request->header( @$_ ) for pairs @$headers;
+   }
+   elsif( $headers and reftype $headers eq "HASH" ) {
+      $request->header( $_, $headers->{$_} ) for keys %$headers;
+   }
+
    my ( $user, $pass );
 
    if( defined $uri->userinfo ) {
